-- matrix.e

function abs(atom this)
    atom that
    that = this
    if this < 0 then
       that = 0 - this
    end if
    return that
end function

function is_vector(sequence this)
    integer bool
    bool = 1
    for i = 1 to length(this) do
        if not atom(this[i]) then
            bool = 0
            exit            
        end if
    end for
    return bool
end function

function is_matrix(sequence this)
    integer bool,len1
    bool = 1
    if is_vector(this) then
        return bool
    end if
    len1 = length(this[1])
    for i = 1 to length(this) do
        if length(this[i]) != len1 then
            bool = 0
            exit
        end if
        if not is_vector(this[i]) then
            bool = 0
            exit
        end if
    end for
    return bool
end function

global type vector(sequence this)
    return is_vector(this)
end type

global type matrix(sequence this)
    return is_matrix(this)
end type

function sum(vector this)
    atom total
    total = 0
    for i = 1 to length(this) do
        total += this[i]
    end for
    return total
end function

function vslice(matrix this,integer column)
    vector s
    s = {}
    for i = 1 to length(this) do
        s &= this[i][column]
    end for
    return s
end function

global function transpose(matrix this)
    matrix t
    t = {}
    if is_vector(this) then
        for i = 1 to length(this) do
            t &= {{this[i]}}
        end for
    else
        for i = 1 to length(this[1]) do
            t &= {vslice(this,i)}
        end for
    end if
    return t
end function

global function multiply(matrix this,matrix that)
    matrix result
    vector line
    result = {}
    line = {}
    if is_vector(this) then
        if length(this) != length(vslice(that,1)) then
            puts(2,"Matrices cannot by multiplied")
            return result
        end if
        for  i = 1 to length(that) do
            line &= sum(this*vslice(that,i))
        end for
        result &= line
    else
        if length(this[1]) != length(vslice(that,1)) then
            puts(2,"Matrices cannot be multiplied")
            return result
        end if
        for i = 1 to length(this) do
            for j = 1 to length(that[1]) do
                line &= sum(this[i]*vslice(that,j))
            end for
            result &= {line}
            line = {}
        end for
    end if
    return result
end function

global function inverse(matrix this)
    atom BIG_ENTRY,EPSILON,TEMP
    matrix B,that
    vector L,M
    integer cols,rows,I_PIVOT, J_PIVOT
    cols = length(this[1])
    rows = length(this)
    L = repeat(0,rows)
    M = L
    B = this
    that = {}
    if rows != cols then
        puts(2,"** Input matrix is not square **\n")
        return {}
    end if
    -- initiate the row and column interchange information
    for k = 1 to rows do
        L[k] = k   -- row interchange information
        M[k] = k   -- column interchange information
    end for

    -- major loop for inverse
    for k = 1 to rows do
        -- & search for row and column index I_PIVOT, J_PIVOT
        -- & both in (k .. B'LAST(1) ) for maximum B(I,J)
        -- & in absolute value :BIG_ENTRY
        BIG_ENTRY = 0.0
        -- check matrix singularity
        for i = k to rows do
        	for j = k to rows do
                if abs(B[i][j]) > abs(BIG_ENTRY) then
                    BIG_ENTRY = B[i][j]
                    I_PIVOT = i
                    J_PIVOT = j
                end if
        	end for
        end for

    	if BIG_ENTRY = 0.0 then
            puts(2,"** Input matrix is Singular **\n")
            return {}
    	else
            EPSILON = rows * (abs(BIG_ENTRY)) * 0.000001
    	end if
    	if abs(BIG_ENTRY) < EPSILON then
            puts(2,"** Input matrix is Singular **\n")
            return {}
        end if
        -- interchange row and column
        --& interchange k-th and I_PIVOT-th rows
        if I_PIVOT != k then
            for j = 1 to rows do
                TEMP = B[I_PIVOT][j]
                B[I_PIVOT][j] = B[k][j]
                B[k][j] = TEMP
            end for
            L[k] = I_PIVOT
        end if
       --& interchange k-th and J_PIVOT-th columns
        if J_PIVOT !=k then
        	for i = 1 to rows do
                TEMP = B[i][J_PIVOT]
                B[i][J_PIVOT] = B[i][k]
                B[i][k] = TEMP
        	end for
        	M[k] = J_PIVOT
        end if

        --& divide i-th column by minus pivot (-BIG_ENTRY)
        
        for i = 1 to rows do
            if i != k then
                B[i][k] = B[i][k] / (-BIG_ENTRY)
            end if
        end for

        -- reduce matrix row by row
        
        for i = 1 to rows do
            if i != k then
                for j = 1 to rows do
                    if j != k then
                        B[i][j] = B[i][j] + B[i][k] * B[k][j]
                    end if
                end for
            end if
        end for

        --& divide k-th row by pivot
        
        for j = 1 to rows do
            if j != k then
                B[k][j] = B[k][j] / BIG_ENTRY
            end if
        end for
        B[k][k] = 1.0 / BIG_ENTRY

    end for   -- end of major inverse loop

    -- final column and row interchange to obtain
    -- inverse of A, i.e. A**(-1)
--
    for k = rows to 1 by -1 do
        -- column interchage
        J_PIVOT = L[k]
        if J_PIVOT != k then
	       --& interchange B(i,J_PIVOT) and B(i,k) for each row i
            for i = 1 to rows do
                TEMP = B[i][J_PIVOT]
                B[i][J_PIVOT] = B[i][k]
                B[i][k] = TEMP
            end for
        end if
        -- row interchage
        I_PIVOT = M[k]
        if I_PIVOT != k then
            --& INTECHANGE B(I_PIVOT,j) and B(k,j) for each column j
            for j = 1 to rows do
                TEMP = B[I_PIVOT][j]
                B[I_PIVOT][j] = B[k][j]
                B[k][j] = TEMP
            end for
        end if
    end for
    
    -- to here

    -- for testing
    that = B
    return that
end function