--<HTML>
--<HEAD>
--<Title>Library: stats</TITLE>
--<LINK HREF=eustyle.css REL="STYLESHEET" TYPE="text/css">
--</HEAD>
--<BODY>
--<div>
--<font color = teal>
--<div><a href=>stats
-- 1.2.1</a></div>
--<div>Statistics library for use with the
--<a href="http://www.rapideuphoria.com/">Euphoria Programming Language</a>
--</div>
--</font>
--</div>

--<h1>stats.e</h1>

-- This stats library providing a number of basic statistical functions.
-- The definition of each function is given with the details of the function.

--<h3>Preparing the data</h3>

-- All functions take as parameters one or more objects containing arrays
-- of data.
--<p>
-- Consequently in order to use these functions you need to arrange your data
-- in an array, within a sequence.
-- Sometimes, however, your data may be in the form of a frequency
-- distribution - pairs of values and frequencies.
-- In this case you can easily use
-- Euphoria's 'repeat' function and the concatenation operator '&' to create
-- the right form of input.
--<p>
-- N.B. No provision is made in this library for 'missing values'.
-- If you wish to use these functions then you must remove such data from
-- your sequence.
--<p>
-- Most functions operate on integer or real data values without problem.
-- Some functions, however, have two forms:
--<ul>
--<li> one for integer data, and
--<li> one for real data.
--</ul>
-- See <a href = #dual>Dual functions</a> for details of the exceptions.
--<p>
-- The function <font color=purple>frequency</font> can handle string-type
-- data as well as numeric data.

--<h3><a name=moments>Moments</a></h3>
-- Moments in statistics are defined in a systemmatic way.
-- Each value can be seen as a deviation from a given value.
-- The mean of these deviations, raised to the power 'r' is known as the
-- rth moment of the distribution, or sometimes the moment of order r.
--<p>
-- A number of the functions in this library are defined using moments.
-- Mainly for this reason all the functions are based on 'n' (the number of
-- observations) and not on 'n-1' (for sample-bound estimates). If you wish to
-- record such estimates then you can easily make the simple calculation
-- adjustment before reporting your results.

--<h2>Library specification</h2>
--<h3>Version: 1.2.1</h3>
--<dir>
--    Date: February 2010
--<br>Author: C A Newbould (canewbould@users.sourceforge.net)
--<br>Licence: Free
--<br> Changes:
--<ul>
--<li>Version 1.2.1
--<ul>
--<li> added the <font color=purple>zeroiseRoot</font> function, so that
--     more than one frequency Tree can be analysed at a time.
--</ul>
--<li>Version 1.2.0
--<ul>
--<li> added the <font color=purple>anova</font> function
--<li> added the <font color=purple>count</font> function
--<li> added the simple linear <font color=purple>regression</font> function
--<li> made the <font color=purple>sum</font> function global
--<li> added the <font color=purple>t_test</font> function
--</ul>
--<li>Version 1.1.0
--<ul>
--<li> added the <font color=purple>frequency</font> function and associated
-- supporting data and routines
--<li> added a function <font color=purple>printTree</font> which outputs the
-- details of a tree generated by the <font color=purple>frequency</font>
-- function.
--<br> (These functions have been modified from the RDS demo file "tree.ex".)
--</ul>
--<li> Version 1.0.0 is the one filed in the
--<a href=http://www.rapideuphoria.com/stats.zip>Euphoria Archive</a>.
--</ul>
--</dir>

--<h2>Interface</h2>

--<h3>Globally-defined includes</h3>
--<ul>
--<li>sort.e - for function <code>sort</code>
include sort.e
--</ul>

  --<h3>Constants:</h3>
  -- pointers within a Tree
constant NAME = 1,
         COUNT = 2,
         LEFT = 3,
         RIGHT = 4

  -- Local variables (for the entire file)
sequence root root = {}  -- needed for insert and tree routines

  --<h3>Routines:</h3>

  -- local routines
  -- function abs(atom a)
  -- Returns the absolute value.
function abs(atom a)
    if a < 0 then
        return -a
    else
        return a
    end if
end function

  -- local routines
  -- function classify(sequence data, sequence classes)
  -- Classifies each value in the data vector according to the
  -- corresponding class value.
  -- The return value is a matrix of data values, grouped by class values.
function classify(sequence data,sequence classes)
    integer l1,l2,found
    sequence matrix,vector
    l1 = length(data)
    l2 = length(classes)
    if l1=l2 then
        vector = {classes[1]}
        matrix = {{data[1]}}
        for i = 2 to l2 do
            found = find(classes[i],vector)
            if found then
                matrix[found] &= data[i]
            else
                vector &= classes[i]
                matrix &= {{data[i]}}
            end if
        end for
    else
        puts(2,"** ERROR ** - arrays not of same length!\n")
        matrix = {}
    end if
    return matrix
end function

  -- local routines
  -- function insert(sequence node, object data)
  -- Inserts a tree node or increments the count of an existing one.
  -- The return value is the updated tree.
function insert(sequence node, object data)
    integer x
    if length(node) = 0 then
        -- couldn't find it - create new node
        node = {data, 1, {}, {}}
    else
        x = compare(data, node[NAME])
        if x = 0 then
            node[COUNT] += 1
        elsif x < 0 then
            node[LEFT] = insert(node[LEFT], data)     -- insert it on the left
        else
            node[RIGHT] = insert(node[RIGHT], data)   -- insert it on the right
        end if
    end if
    return node
end function

  -- function is_vector(sequence v)
  -- defines the vector type
function is_vector(sequence v)
  integer state
  state = 1
  for i = 1 to length(v) do
    if not atom(v[i]) then
      state = 0
    end if
  end for
  return state
end function

  -- consequential local type
type vector(object v)
    return is_vector(v)
end type

  -- function percentile(vector this,integer p)
  -- Calculates the pth percentile of the vector.
  -- If there are an even number then we take the average of the two adjacent ones.
  -- The return value is the (real) percentile.
  -- Calls external function "sort".
function percentile(vector this,integer p)
    sequence sorted_this
    integer mid,
            n
    sorted_this = sort(this)
    n = length(this)
    mid = floor((p * n)/100)
    if sorted_this[mid] = sorted_this[mid + 1] then
        return sorted_this[mid]
    else
        return (sorted_this[mid] + sorted_this[mid + 1])/2
    end if
end function

  -- function sign(atom a)
  -- Returns the sign of the value (1 or -1).
function sign(atom a)
    atom sign
    if a < 0 then
        sign = -1
    else
        sign = 1
    end if
    return sign
end function

  -- function square(atom a)
  -- Calculates the square of the value.
  -- Calls external built-in "power".
function square(atom a)
    return power(a,2)
end function

  -- needed here - documentation below
global function sum(vector this)
    atom total
    total = 0
    for i = 1 to length(this) do
            total += this[i]
    end for
    return total
end function

  -- function sumOfSquares(vector this)
  -- Calculates the sum of the squares of the values in the called vector.
  -- The return value is the (real) sum of squares of the values in the vector.
global function sumOfSquares(vector this)
        return sum(power(this,2))
end function

  -- function sumOfProducts(vector this, vector that)
  -- Calculates the sum of the products of the values in the called vectors.
  -- The return value is the (real) sum of products of the values in the two
  -- vectors.
function sumOfProducts (vector this, vector that)
        return sum(this * that)
end function

  -- function sumOfCubes(vector this)
  -- Calculates the sum of the cubes of the values in the called vector.
  -- The return value is the (real) sum of cubes of the values in the vector.
function sumOfCubes(vector this)
    return sum(power(this,3))
end function

  --/topic Service Routines
  --/func sumOfFourths(vector this)
  -- Calculates the sum of the fourth power of the values in the
  -- called vector.
  --/n The return value is the (real) sum of fourths of the values in the vector.
function sumOfFourths(vector this)
    return sum(power(this,4))
end function

--<h3>Routines:</h3>
-- ** univariate routines **
--<p>
-- All these functions take, as the parameter, a sequence consisting
-- of atoms in the form of a one-dimensional array. A <code>vector</code> type
-- has been defined with the library to enable type-checking to take place when
-- each function is called.
--<ul>
--<li> [function] <font color=purple>average</font>(<code>vector</code> this)
-- - calculates the first <a href=#moments>moment</a>, centred on zero.
--<br> The return value is the (real) arithmetic average of the values in the
-- vector.

--<li> [function] <font color=purple>count</font>(<code>vector</code> this)
-- - synonym for Euphoria's <code>length</code> function
--<br> The return value is the (real) number of values in the vector.
global function count(vector this)
    return length(this)
end function

global function average(vector this)
    return sum(this)/count(this)
end function

  -- needed here - documentation below
global function variance(vector this)
    return sumOfSquares(this)/count(this) - power(average(this),2)
end function

--<li> [function] <font color=purple>kurtosis</font>(<code>vector</code> this)
-- - calculates the kurtosis of the distribution in the vector.
-- This is based the fourth <a href=#moments>moment</a> about the arithmetic
-- mean.
-- The kurtosis of the distribution is defined as the 'standardised' fourth
-- moment, that is, the fourth moment divided by the fourth power of the
-- standard deviation.
--<br>
-- Because the kurtosis of the normal distribution is 3, some authors subtract
-- the 3 before reporting the value. If you wish to use this convention
-- then simply subtract three from the return value of this function.
--<br> The return value is the (real) kurtosis of the values in the vector.
global function kurtosis(vector this)
    atom sf,
         fourth_moment,
         n
    n = count(this)
    sf = sumOfFourths(this)
       - 4 * average(this) * sumOfCubes(this)
       + 6 * power(average(this),2) * sumOfSquares(this)
       - 4 * power(average(this),3) * sum(this)
       + n * power(average(this),4)
    fourth_moment = sf/n
    return fourth_moment/power(variance(this),2)
end function

--<li> [function] <font color=purple>maximum</font>(<code>vector</code>
-- this)
-- - calculates the maximum value in the vector.
--<br> The return value is the (real) maximum.
global function maximum(vector this)
    atom max
    max = this[1]
    for i = 2 to length(this) do
        if this[i] > max then
            max = this[i]
        end if
    end for
    return max
end function

--<li> [function] <font color=purple>median</font>(<code>vector</code>
-- this)
-- - calculates the median value in the vector.
--<br> If there are an even number then we take the average of the two "middle"
-- ones.
--<br> The return value is the (real) median.
global function median(vector this)
    return percentile(this,50)
end function

--<li> [function] <font color=purple>minimum</font>(<code>vector</code>
-- this)
-- - calculates the minimum value in the vector.
--<br> The return value is the (real) minimum.
global function minimum(vector this)
    atom min
    min = this[1]
    for i = 2 to length(this) do
        if this[i] < min then
            min = this[i]
        end if
    end for
    return min
end function

--<li> [function] <font color=purple>quartile</font>(<code>vector</code>
-- this)
-- Calculates the qth (1st, 2nd or 3rd) quartile of the vector.
--<br> If there are an even number then we take the average of the two adjacent
-- ones.
--<br> The return value is the (real) quartile.
global function quartile(vector this,integer q)
    if q > 0 and q < 4 then
        return percentile(this,25 * q)
    else
        puts(2,"\nOnly values of 1, 2 and 3 allowed!")
        return 0
    end if
end function

--<li> [function] <font color=purple>range</font>(<code>vector</code>
-- this)
-- - calculates the difference between the maximum and minimum value in the
-- vector.
--<br> The return value is the (real) range.
global function range(vector this)
    return maximum(this) - minimum(this)
end function

--<li> [function] <font color=purple>skewness</font>(<code>vector</code>
-- this)
-- - calculates the skewness of the distribution in the vector.
-- This is based the third /moment about the arithmetic mean.
-- The skewness of the distribution is defined as the 'standardised' third
-- <a href=#moments>moment</a>, that is, the third moment divided by the
-- cube of the standard deviation.
--<br>
-- The return value is the (real) skewness coefficient of the values in the
-- vector.
--<br>
-- Assumes no missing values.
global function skewness(vector this)
    atom n,
         numerator,
         denominator
    n = length(this)
    numerator = sumOfCubes(this)
                - 3 * sumOfSquares(this) * average(this)
                + 3 * power(average(this),2) * sum(this)
                - n * power(average(this),3)
    numerator /= n
    denominator = power(sqrt(variance(this)),3)
    return numerator / denominator
end function

  -- necessarily defined above
--<li> [function] <font color=purple>sum</font>(<code>vector</code> this)
-- - calculates the sum of the values in the called vector.
--<br> The return value is the (real) sum of values in the vector.

  -- necessarily defined above
--<li> [function] <font color=purple>variance</font>(<code>vector</code> this)
-- - calculates the second <a href=#moments>moment</a> of the distribution,
-- about the arithmetic mean.
--<br> The return value is the (real) variance of the values in the vector.

--</ul>

-- ** more generalised univariate routine **
--<p>
-- This function take, as the parameter, a sequence consisting
-- of objects in the form of a vector.
--<ul>
--<li> [function] <font color=purple>frequency</font>(<code>sequence</code>
-- this)
-- - determines the frequency distribution of the vector.
--<br> The return value is a tree containing counts.
--<br> Each node contains the value, its count and the pointers to its
--    neighbours.
--<br> Will handle any kind of values in the vector.
global function frequency(sequence this)
	for i = 1 to count(this) do
        root = insert(root, this[i])
    end for
    return root
end function

--<li> [procedure] <font color=purple>printTree</font>(<code>sequence</code>
-- node)
-- - displays the counts in a tree structure.
global procedure printTree(sequence node)
    if length(node) = 0 then
        return
    end if
    printTree(node[LEFT])
    if sequence(node[NAME]) then
        printf(1, "Value: %s; frequency: %d\n", {node[NAME], node[COUNT]})
    else
        printf(1, "Value: %d; frequency: %d\n", {node[NAME], node[COUNT]})
    end if
    printTree(node[RIGHT])
end procedure

--<li> [procedure] <font color=purple>zeroiseRoot</font>()
-- - zeroises a tree structure (if more than one frequency Tree is to be
--   analysed).
global procedure zeroiseRoot()
    root = {}
end procedure

--</ul>

-- ** bivariate routines **
--<p>
--N.B. All these functions take, as parameters, two sequences consisting
-- of atoms in the form of a vector. A <code>vector</code> type
-- has been defined with the library to enable type-checking to take place when
-- each function is called. The data are analysed pair-wise: that is, the first
-- value in vector1 is associated with the first value in vector2, etc.
--<ul>

global function anova(vector this, vector class)
    integer n, nclasses
    atom sserror, ssmethods, sstotal
    sequence data
    n = count(this)
    sstotal = variance(this)*n
    sserror = 0
    data = classify(this,class)
    nclasses = length(data)
    for i = 1 to nclasses do
        sserror += variance(data[i])*count(data[i])
    end for
    ssmethods = sstotal - sserror
    return {{ssmethods,sserror},
            {nclasses-1,n-nclasses}
           }
end function

--<li> [function] <font color=purple>anova</font>(<code>vector</code>
-- this, <code>vector</code> class)
-- - calculates the <a href=http://en.wikipedia.org/wiki/One-way_ANOVA>one-way
-- analysis of variance</a> of the array, given the
-- class groupings.
--<br> The return value is a sequence containing the following arrays:
--<ul>
--<li> the sum of squares due to methods (classification) and error
--<li> the corresponding degrees of freedom
--</ul>
-- The total sum of squares is the sum of the values in the first array and
-- the total degrees of freedom are the sum of the values in the second array.

  -- needed here - documentation below
global function covariance(vector this, vector that)
    return sumOfProducts(this,that)/count(this)
         - average(this)* average(that)
end function

--<li> [function] <font color=purple>correlation</font>(<code>vector</code>
-- this, <code>vector</code> that)
-- - calculates the correlation coefficient of the paired values in the two
-- vectors.
--<br> The return value is the (real) correlation of the values in the vectors.
global function correlation(vector this, vector that)
    return covariance(this, that)/sqrt(variance(this)*variance(that))
end function

  -- necessarily defined above
--<li> [function] <font color=purple>covariance</font>(<code>vector</code>
-- this, <code>vector</code> that)
-- - calculates the covariance of the paired values in the two vectors.
--<br> The return value is the (real) covariance of the values in the vectors.

--<li> [function] <font color=purple>regression</font>(<code>vector</code>
-- this,<code>vector</code> that)
-- - calculates the simple linear regression coefficient of the values in
-- this given the values in that.
--<br> The return value is the (real) simple linear regression coefficient.
global function regression(vector this, vector that)
    return covariance(this, that)/variance(that)
end function

--<li> [function] <font color=purple>t_test</font>(<code>vector</code>
-- this,<code>vector</code> that)
-- - calculates the <a href=http://www.socialresearchmethods.net/kb/stat_t.php>
--t-test</a> for the difference between the means of the two vectors.
--<br> The return value is a <code>sequence</code> giving:
--<ul>
--<li>the (real) t statistic, with sign
--<li>a sequence containing the degrees of freedom (1,df)
--</ul>
-- The statistical significance of the result can be looked up in the tables
-- in the library F.e.
global function t_test(vector this,vector that)
    atom n1,n2,S1,S2,df,L1,L2,sig,xt2,t
    n1 = count(this)
    S1 = sum(this)

    n2 = count(that)
    S2 = sum(that)

    df = n1 + n2 -2

    L1 = n1*sumOfSquares(this) - square(S1)
    L2 = n2*sumOfSquares(that) - square(S2)

    sig = sign(average(this) - average(that))

    xt2 = df * square(count(that)*sum(this) - count(this)*sum(that))/
          ((df + 2)* (count(that)*L1 + count(this)*L2))
    t = sqrt(abs(xt2))
    return {sig*t,{1,df}}
end function
--</ul>

-- ** <a name=dual>Dual functions</a> **
--<p> Instances where there are separate functions for integer and real data.
--<p>
--<TABLE BORDER=2>
--<TR> <TD>&nbsp;</TD>
--     <TH>Distributions</TH>
--     <TH>Modal value[s]</TH>
--     </TR>

--<TR> <TH>Integer</TH>
--     <TD><font color=purple>distribution</font></TD>
--     <TD><font color=purple>mode</font></TD>

--<TR> <TH>Real</TH>
--     <TD><font color=purple>real_distribution</font></TD>
--     <TD><font color=purple>real_mode</font></TD>

--</TABLE>

--<h4>Integer functions</h4>
--<ul>
--<li> [function] <font color=purple>distribution</font>(<code>vector</code>
-- this)
-- - determines the frequency distribution of the vector.
--<br> The base unit is relative to the minimum.
--<br> The return value is a sequence with counts.
-- It consists of the base unit, followed by all the counts.
global function distribution(vector this)
    atom base,
         min,
         val
    sequence dist
    min = minimum(this)
    if min >= 0 then
        base = min
    else
        base = 0
    end if
    dist = repeat(0,range(this) + 1)
    for i = 1 to count(this) do
        val = 1 + this[i] - min
        dist[val] += 1
    end for
    return base & dist
end function
--<li> [function] <font color=purple>mode</font>(<code>vector</code>
-- this)
-- - determines the modal value(s) of the frequency distribution of the
-- vector.
--<br> The return value is a sequence with one or more values.
global function mode(vector this)
    atom max
    sequence dist,
             modes
    dist = distribution(this)
    max = maximum(dist[2..$])
    modes = {}
    for i = 2 to length(dist) do
        if dist[i] = max then
            if dist[1] > 0 then
                modes &= i + dist[1] - 2
            else
                modes &= i + minimum(this) - 2
            end if
        end if
    end for
    return modes
end function

--</ul>

--<h4>Real functions</h4>
--<ul>
--<li> [function]
-- <font color=purple>real_distribution</font>(<code>vector</code>
-- this,<code>integer</code> interval)
-- - determines the frequency distribution of the vector.
--<br> The base unit is relative to the minimum.
--<br> The intervals used in the distribution are determined by the second
--     parameter - starting with the integer immediately less than
--     the minimum value.
--<br> The return value is a sequence with counts.
--     Its value consists of the base unit, followed by all the counts.
--<br> NB. The base will always be an integer.

global function real_distribution(vector this,integer interval)
    atom base,
         ints,
         min,
         val
    sequence dist
    min = floor(minimum(this))
    if min >= 0 then
        base = min
    else
        base = 0
    end if
    ints = floor((range(this) + 1)/interval) + 1
    dist = repeat(0,ints)
    for i = 1 to count(this) do
        val = floor((interval + this[i] - min)/interval)
        dist[val] += 1
    end for
    return base & dist
end function

--<li> [function]
-- <font color=purple>real_mode</font>(<code>vector</code>
-- this,<code>integer</code> interval)
-- - determines the modal value(s) of the frequency distribution of the vector.
--<br> The return value is a sequence with one or more values.
--<br> Assumes that the values in the array will be grouped into intervals
--     - of size 'interval'.
global function real_mode(vector this,integer interval)
    atom max
    sequence dist,
             modes
    dist = real_distribution(this,interval)
    max = maximum(dist[2..$])
    modes = {}
    for i = 2 to length(dist) do
        if dist[i] = max then
            if dist[1] > 0 then
                modes &= i + dist[1] - 2
            else
                modes &= i + minimum(this) - 2
            end if
        end if
    end for
    return modes
end function

--</ul>

--</body>
--</HTML>